import json
from typing import Literal

from pydantic import Field, ImportString
from visivo.models.destinations.destination import Destination
from visivo.models.test_run import TestRun


class SlackDestination(Destination):
    """
    You can configure slack alerts by setting up an incoming message slack webhook. Once you do that, the set up in Visivo is super simple:
    ``` yaml
    alerts:
      - name: slack-destination #any name you choose
        type: slack
        webhook_url: {% raw %}{{ env_var("SLACK_WEBHOOK")}}{% endraw %}
    ```
    """

    webhook_url: str = Field(
        None,
        description="An incoming message slack webhook url. You can set one of those up by following <a href='https://api.slack.com/messaging/webhooks'>these instructions</a>.",
    )
    type: Literal["slack"] = Field(
        description="The type of Destination Destination. Needs to be `slack` to configure a slack destination",
    )

    def alert(self, test_run: TestRun):
        import requests

        json_headers = {"content-type": "application/json"}

        if not test_run.failures:
            return

        body = {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": f"There were test failures running against {test_run.target_name}",
                    },
                }
            ]
        }

        test_failures = ""
        for test_failure in test_run.failures:
            test_failures += f"* {test_failure.test_id} - {test_failure.message}\n"

        body["blocks"].append(
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": f"{test_failures}",
                },
            }
        )

        requests.post(self.webhook_url, data=json.dumps(body), headers=json_headers)



from typing import Union
from typing_extensions import Annotated
from pydantic import Field

from visivo.models.destinations.console_destination import ConsoleDestination
from visivo.models.destinations.email_destination import EmailDestination
from visivo.models.destinations.slack_destination import SlackDestination

DestinationField = Annotated[
    Union[SlackDestination, EmailDestination, ConsoleDestination],
    Field(discriminator="type"),
]



def run_phase(
    default_source: str,
    output_dir: str,
    working_dir: str,
    name_filter: str = None,
    run_only_changed: bool = False,
    threads: int = None,
    soft_failure=False,
    dbt_profile: str = None,
    dbt_target: str = None,
):
    from visivo.logging.logger import Logger
    from visivo.query.runner import Runner
    from visivo.commands.compile_phase import compile_phase

    project = compile_phase(
        default_source=default_source,
        working_dir=working_dir,
        output_dir=output_dir,
        name_filter=name_filter,
        dbt_profile=dbt_profile,
        dbt_target=dbt_target,
    )
    if threads is None and project.defaults and project.defaults.threads:
        threads = project.defaults.threads
    elif threads is None:
        threads = 8
    else:
        threads = int(threads)

    source_details = (
        "\n" if default_source == None else f" and default source {default_source}\n"
    )
    Logger.instance().info(f"Running project across {threads} threads" + source_details)

    runner = Runner(
        project=project,
        output_dir=output_dir,
        threads=threads,
        soft_failure=soft_failure,
        run_only_changed=run_only_changed,
        name_filter=name_filter,
    )
    runner.run()
    return runner



from visivo.models.base.parent_model import ParentModel
from visivo.models.dag import all_descendants_of_type
from visivo.models.models.local_merge_model import LocalMergeModel
from visivo.models.models.model import Model
from visivo.models.project import Project
from visivo.query.jobs.job import (
    Job,
    JobResult,
    format_message_failure,
    format_message_success,
    start_message,
)
from time import time
from visivo.logging.logger import Logger


def action(local_merge_model: LocalMergeModel, output_dir, dag):
    Logger.instance().info(start_message("LocalMergeModel", local_merge_model))
    try:
        start_time = time()
        local_merge_model.insert_dependent_models_to_sqlite(
            output_dir=output_dir, dag=dag
        )
        success_message = format_message_success(
            details=f"Updated data for model \033[4m{local_merge_model.name}\033[0m",
            start_time=start_time,
            full_path=local_merge_model.get_sqlite_source(
                output_dir=output_dir, dag=dag
            ).database,
        )
        return JobResult(success=True, message=success_message)
    except Exception as e:
        failure_message = format_message_failure(
            details=f"Failed query for model \033[4m{local_merge_model.name}\033[0m",
            start_time=start_time,
            full_path=local_merge_model.get_sqlite_source(
                output_dir=output_dir, dag=dag
            ).database,
            error_msg=str(repr(e)),
        )
        return JobResult(success=False, message=failure_message)


def jobs(dag, output_dir: str, project: Project, name_filter: str):
    local_merge_models = all_descendants_of_type(
        type=LocalMergeModel, dag=dag, from_node=project
    )

    if name_filter:
        included_nodes = project.nodes_including_named_node_in_graph(name=name_filter)
    else:
        included_nodes = project.descendants()
    local_merge_models = set(local_merge_models).intersection(included_nodes)

    jobs = []
    for local_merge_model in local_merge_models:
        jobs.append(
            Job(
                item=local_merge_model,
                source=local_merge_model.get_sqlite_source(
                    output_dir=output_dir, dag=dag
                ),
                action=action,
                local_merge_model=local_merge_model,
                output_dir=output_dir,
                dag=dag,
            )
        )
    return jobs



from typing import Literal, Optional
from visivo.models.sources.sqlalchemy_source import SqlalchemySource
from click import ClickException
from pydantic import Field

SnowflakeType = Literal["snowflake"]


class SnowflakeSource(SqlalchemySource):
    """
    SnowflakeSources hold the connection information to Snowflake data sources.

    !!! example

        === "Simple"

            ``` yaml
                sources:
                  - name: snowflake_source
                    type: snowflake
                    database: DEV
                    warehouse: DEV
                    account: ab12345.us-west-1.aws
                    db_schema: DEFAULT
                    username: {% raw %}{{ env_var('SNOWFLAKE_USER') }}{% endraw %}
                    password: {% raw %}{{ env_var('SNOWFLAKE_PASSWORD') }}{% endraw %}
            ```

    Note: Recommended environment variable use is covered in the [sources overview.](/topics/sources/)
    """

    account: Optional[str] = Field(
        None,
        description="The snowflake account url. Here's how you find this: [snowflake docs](https://docs.snowflake.com/en/user-guide/admin-account-identifier).",
    )
    warehouse: Optional[str] = Field(
        None,
        description="The compute warehouse that you want queries from your Visivo project to leverage.",
    )
    role: Optional[str] = Field(
        None,
        description="The access role that you want to use when running queries.",
    )
    timezone: Optional[str] = Field(
        None,
        description="The timezone that you want to use by default when running queries.",
    )

    type: SnowflakeType
    connection_pool_size: Optional[int] = Field(
        8, description="The pool size that is used for this connection."
    )

    def get_dialect(self):
        return "snowflake"

    def url(self):
        from snowflake.sqlalchemy import URL
        url_attributes = {
            "user": self.username,
            "password": self.get_password(),
            "account": self.account,
        }
        # Optional attributes where if its not set the default value is used
        if self.timezone:
            url_attributes["timezone"] = self.timezone
        if self.warehouse:
            url_attributes["warehouse"] = self.warehouse
        if self.role:
            url_attributes["role"] = self.role
        if self.database:
            url_attributes["database"] = self.database
        if self.db_schema:
            url_attributes["schema"] = self.db_schema

        url = URL(**url_attributes)

        return url


import click
from .options import (
    output_dir,
    working_dir,
    source,
    name_filter,
    dbt_profile,
    dbt_target,
)


@click.command()
@source
@working_dir
@output_dir
@name_filter
@dbt_profile
@dbt_target
def compile(working_dir, output_dir, source, name_filter, dbt_profile, dbt_target):
    """
    Parses the files in your working directory, extracting visivo configurations and then using those configurations to build the trace queries and a project.json file in your source directory. Queries are not run on compile, just written.
    """
    from visivo.logging.logger import Logger

    Logger.instance().debug("Compiling")

    from visivo.commands.compile_phase import compile_phase

    compile_phase(
        default_source=source,
        working_dir=working_dir,
        output_dir=output_dir,
        name_filter=name_filter,
        dbt_profile=dbt_profile,
        dbt_target=dbt_target,
    )
    Logger.instance().success("Done")



# supports more generic connections than the snowflake specific connector
from visivo.models.alert import Alert
from visivo.models.base.parent_model import ParentModel
from visivo.models.dag import all_descendants_of_type
from visivo.models.project import Project
from visivo.models.test import Test
from visivo.models.trace import Trace
from visivo.models.test_run import TestRun, TestFailure, TestSuccess
from visivo.models.destinations.destination import Destination
from typing import Any, List
import json
import click
from datetime import datetime
import warnings


warnings.filterwarnings("ignore")


class Runner:
    def __init__(
        self,
        tests: List[Test],
        project: Project,
        output_dir: str,
        dag: Any,
    ):
        self.tests = tests
        self.project = project
        self.output_dir = output_dir
        self.dag = dag

    def run(self):
        test_run = TestRun()
        for test in self.tests:
            for assertion in test.assertions:
                try:
                    passed = assertion.evaluate(
                        dag=self.dag, project=self.project, output_dir=self.output_dir
                    )
                    if passed:
                        click.echo(click.style(".", fg="green"), nl=False)
                        success = TestSuccess(test_id=test.path)
                        test_run.add_success(success=success)
                    else:
                        click.echo(click.style("F", fg="red"), nl=False)
                        failure = TestFailure(
                            test_id=test.path, message=assertion.value
                        )
                        test_run.add_failure(failure=failure)
                except Exception as e:
                    failure = TestFailure(test_id=test.path, message=str(e))
                    click.echo(click.style("F", fg="red"), nl=False)
                    test_run.add_failure(failure=failure)
        test_run.finished_at = datetime.now()
        click.echo("")
        click.echo(test_run.summary())

        alerts = all_descendants_of_type(type=Alert, dag=self.dag)

        for alert in set(alerts):
            if alert.if_ and alert.if_.evaluate(
                dag=self.dag,
                project=self.project,
                output_dir=self.output_dir,
                test_run=test_run,
            ):
                alert.alert(test_run=test_run)

        return test_run



import os
import json

from visivo.discovery.discover import Discover
from visivo.models.dag import all_descendants_of_type
from visivo.models.defaults import Defaults
from visivo.models.models.csv_script_model import CsvScriptModel
from visivo.models.sources.source import Source
from visivo.models.models.model import Model
from visivo.models.base.parent_model import ParentModel
from visivo.parsers.serializer import Serializer
from visivo.query.query_string_factory import QueryStringFactory
from visivo.query.trace_tokenizer import TraceTokenizer
from visivo.query.query_writer import QueryWriter
from visivo.logging.logger import Logger
from .utils import parse_project_file
from .dbt_phase import dbt_phase


def compile_phase(
    default_source: str,
    working_dir: str,
    output_dir: str,
    name_filter: str = None,
    dbt_profile: str = None,
    dbt_target: str = None,
):
    dbt_phase(working_dir, output_dir, dbt_profile, dbt_target)

    Logger.instance().debug("Compiling project")
    project = parse_project_file(working_dir, output_dir, default_source)

    os.makedirs(output_dir, exist_ok=True)
    with open(f"{output_dir}/project.json", "w") as fp:
        serializer = Serializer(project=project)
        fp.write(
            serializer.dereference().model_dump_json(exclude_none=True, by_alias=True)
        )

    dag = project.dag()
    for trace in project.filter_traces(name_filter=name_filter):
        model = all_descendants_of_type(
            type=Model, dag=dag, from_node=trace
        )[0]
        if isinstance(model, CsvScriptModel):
            source = model.get_sqlite_source(output_dir=output_dir)
        else:
            source = all_descendants_of_type(
                type=Source, dag=dag, from_node=model
            )[0]
        tokenized_trace = TraceTokenizer(
            trace=trace, model=model, source=source
        ).tokenize()
        query_string = QueryStringFactory(tokenized_trace=tokenized_trace).build()
        QueryWriter(
            trace=trace, query_string=query_string, output_dir=output_dir
        ).write()

    Logger.instance().debug("Project compiled")
    return project



from visivo.models.base.named_model import NamedModel
from visivo.models.selector import Selector
from .base.base_model import generate_ref_field
from .base.parent_model import ParentModel
from pydantic import Field
from typing import Optional, Literal
from .chart import Chart
from .table import Table
from pydantic import model_validator


class Item(NamedModel, ParentModel):
    """
    The Item houses a single chart, table, selector or markdown object.

    It also informs the width that the chart, table or markdown should occupy within a row. Widths are evaluated for each item in a row relative to all of the other items in the row.

    In the example below, the markdown would take up 1/4th of the row and would be positioned on the left edge. The table would also take up 1/4th of the page and would sit to the right of the markdown. The chart would take up 1/2 of the page and would touch the right edge of the row.

    ``` yaml
    items:
      - width: 1
        markdown: "# Some inline **markdown**"
      - width: 1
        table: ref(table-name)
      - width: 2
        chart: ref(chart-name)
      - width: 1
        selector: ref(selector-name)
    ```
    ## Markdown
    You can use markdown to add formatted text to your dashboard. Visivo markdown supports [CommonMark](https://commonmark.org/help/) and [GitHub Flavored Markdown](https://github.github.com/gfm/). You can also 
    render raw HTML within your markdown.

    To control the alignment of markdown content, you can use the `align` and `justify` properties.
    === "Horizontal Alignment (align)"
        Controls how text aligns horizontally within the container:

        `align: left` (default)
        ```
        [Header     ]
        [Paragraph  ]
        [List       ]
        ```

        `align: center`
        ```
        [  Header   ]
        [ Paragraph ]
        [   List    ]
        ```

        `align: right`
        ```
        [     Header]
        [  Paragraph]
        [      List]
        ```

    === "Vertical Distribution (justify)"
        Controls how content blocks are distributed vertically in fixed-height containers:

        `justify: start` (default)
        ```
        [Header     ]
        [Paragraph  ]
        [List       ]
        [           ]
        [           ]
        ```

        `justify: center`
        ```
        [           ]
        [Header     ]
        [Paragraph  ]
        [List       ]
        [           ]
        ```

        `justify: between`
        ```
        [Header     ]
        [           ]
        [Paragraph  ]
        [           ]
        [List       ]
        ```

        `justify: around`
        ```
        [           ]
        [Header     ]
        [           ]
        [Paragraph  ]
        [           ]
        [List       ]
        [           ]
        ```

        `justify: evenly`
        ```
        [           ]
        [Header     ]
        [Paragraph  ]
        [List       ]
        [           ]
        ```

        `justify: end`
        ```
        [           ]
        [           ]
        [Header     ]
        [Paragraph  ]
        [List       ]
        ```
    """

    width: int = Field(
        1,
        description="The width of the Item determines is evaluated relative to the other items in a row.",
    )
    markdown: Optional[str] = Field(
        None, description="Markdown text to include in the dashboard."
    )
    align: Optional[Literal["left", "center", "right"]] = Field(
        None, 
        description="Alignment of markdown content. Only valid when markdown is set. Options are 'left', 'center', or 'right'."
    )
    justify: Optional[Literal["start", "end", "center", "between", "around", "evenly"]] = Field(
        None,
        description="Justification of markdown content within its container. Options are 'start', 'end', 'center', 'between', 'around', or 'evenly'."
    )
    chart: Optional[generate_ref_field(Chart)] = Field(
        None, description="A chart object defined inline or a ref() to a chart."
    )
    table: Optional[generate_ref_field(Table)] = Field(
        None, description="A Table object defined inline or a ref() to a table"
    )
    selector: Optional[generate_ref_field(Selector)] = Field(
        None, description="A Selector object defined inline or a ref() to a selector"
    )

    @model_validator(mode="before")
    @classmethod
    def validate_unique_item_types(cls, data: any):
        markdown, chart, table, selector = (
            data.get("markdown"),
            data.get("chart"),
            data.get("table"),
            data.get("selector"),
        )
        items_set = [i for i in [markdown, chart, table, selector] if i is not None]
        if len(items_set) > 1:
            raise ValueError(
                'only one of the "markdown", "chart", "table", or "selector" properties should be set on an item'
            )
        return data

    @model_validator(mode='before')
    @classmethod
    def validate_align_with_markdown(cls, data: any):
        align = data.get('align')
        markdown = data.get('markdown')
        if markdown is not None and align is None:
            data['align'] = 'left'
        elif align is not None and markdown is None:
            raise ValueError("The 'align' property can only be set when 'markdown' is present in the same item")
        return data

    @model_validator(mode='before')
    @classmethod
    def validate_justify_with_markdown(cls, data: any):
        justify = data.get('justify')
        markdown = data.get('markdown')
        if markdown is not None and justify is None:
            data['justify'] = 'start'
        elif justify is not None and markdown is None:
            raise ValueError("The 'justify' property can only be set when 'markdown' is present in the same item")
        return data

    def child_items(self):
        child = self.__get_child()
        if child:
            return [self.__get_child()]
        return []

    def __get_child(self):
        if self.table is not None:
            return self.table
        if self.chart is not None:
            return self.chart
        if self.selector is not None:
            return self.selector



from typing import Optional
from pydantic import BaseModel, Field


class Dbt(BaseModel):
    """
    Configuration for pulling models and sources from a dbt project.

    The integration will create models and sources from the dbt project and store them in the output directory.

    It will pull all the "targets" defined in the profiles.yml file as sources.
    Each of these sources will be named in the format "prefix + dbt_profile + dbt_target".

    Then it will create each of the dbt models as Visivo models pointing to the default profile and target.
    These models will be named in the format "prefix + dbt_model".

    Using all the default values:
    ``` yaml
    dbt:
      enabled: true
    ```

    Or specify the input and/or output file and locations:

    ``` yaml
    dbt:
      output_file: includes/dbt.yml
      dbt_project_yml_location: dbt
      profiles_yml_location: dbt
    ```
    """

    enabled: Optional[bool] = Field(
        True,
        description="Whether to enable the dbt phase.  Defaults to true.",
    )
    output_file: Optional[str] = Field(
        None,
        description="The file to store the dbt models and sources relative to the working directory.  Defaults to the '$output_directory/dbt.yml'. It is useful to store the file in a different location so it can be checked into source control.",
    )
    dbt_project_yml_location: Optional[str] = Field(
        None,
        description="Location for the dbt_project.yml file relative to the working directory.  Defaults to the current working directory.",
    )
    profiles_yml_location: Optional[str] = Field(
        None,
        description="Location for the profiles.yml file relative to the working directory.  Defaults to the current working directory.",
    )
    prefix: str = Field(
        "dbt_",
        description="Prefix for the dbt models and sources. Defaults to 'dbt_'.",
    )

    def get_output_file(self, output_dir: str, working_dir: str):
        if self.output_file:
            return f"{working_dir}/{self.output_file}"
        return f"{output_dir}/dbt.yml"



def _generate_sources(profiles, dbt_target, dbt_prefix):
    import click
    from visivo.commands.utils import get_source_types

    source_types = get_source_types()
    sources = []
    target_found = False
    for profile_name, profile_value in profiles.items():
        for target_name, target_value in profile_value["outputs"].items():
            if target_name == dbt_target:
                target_found = True
            source = target_value
            source["name"] = f"{dbt_prefix}{profile_name}_{target_name}"
            if "schema" in source:
                source["db_schema"] = source.pop("schema")
            if "user" in source:
                source["username"] = source.pop("user")
            if "threads" in source:
                source.pop("threads")
            if "type" in source and source["type"] not in source_types:
                raise click.ClickException(
                    f"Target type '{source['type']}' is not supported.  Only {source_types} are supported."
                )
            sources.append(source)

    if not target_found:
        raise click.ClickException(f"Target '{dbt_target}' not found in profiles.yml")

    return sources


def _generate_models(manifest, dbt_profile, dbt_target, dbt_prefix):
    models = []
    if "nodes" in manifest:
        for path, node in manifest["nodes"].items():
            if path.startswith("model."):
                names = []
                if "database" in node:
                    names.append(node["database"])
                if "schema" in node:
                    names.append(node["schema"])
                names.append(node["name"])
                model = {
                    "name": f"dbt_{node['name']}",
                    "sql": "select * from " + ".".join(names),
                    "source": f"ref({dbt_prefix}{dbt_profile}_{dbt_target})",
                }

                models.append(model)

    return models


def dbt_phase(working_dir, output_dir, dbt_profile, dbt_target):
    from visivo.logging.logger import Logger
    from visivo.parsers.parser_factory import ParserFactory
    from visivo.discovery.discover import Discover
    from visivo.models.project import Dbt
    import os
    import click

    discover = Discover(working_dir=working_dir, output_dir=output_dir)
    parser = ParserFactory().build(
        project_file=discover.project_file, files=discover.files
    )
    data = parser.merge_data_files()
    if "dbt" in data and data["dbt"]:
        dbt = Dbt(**data["dbt"])
    else:
        dbt = None

    if dbt and dbt.enabled:
        dbt_root = working_dir
        if dbt.dbt_project_yml_location:
            dbt_root = f"{working_dir}/{dbt.dbt_project_yml_location}"
        dbt_project_file = f"{dbt_root}/dbt_project.yml"
        if not os.path.exists(dbt_project_file):
            raise click.ClickException(
                f"dbt_project.yml file not found at '{dbt_project_file}'"
            )

        profiles_file = f"{working_dir}/profiles.yml"
        if dbt.profiles_yml_location:
            profiles_file = f"{working_dir}/{dbt.profiles_yml_location}/profiles.yml"

        if not os.path.exists(profiles_file):
            raise click.ClickException(
                f"profiles.yml file not found at '{profiles_file}'"
            )

        Logger.instance().debug(f"Found dbt_project file: {dbt_project_file}")
        Logger.instance().debug(f"Found profiles file: {profiles_file}")

        import json
        import ruamel.yaml

        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True

        with open(dbt_project_file, "r") as file:
            dbt_project = yaml.load(file)

        with open(profiles_file, "r") as file:
            profiles = yaml.load(file)

        if not dbt_profile and "profile" in dbt_project:
            dbt_profile = dbt_project["profile"]

        if dbt_profile not in profiles:
            raise click.ClickException(
                f"Profile '{dbt_profile}' not found in profiles.yml"
            )
        if not dbt_target:
            dbt_target = profiles[dbt_profile]["target"]

        manifest_file = f"{dbt_root}/{dbt_project['target-path']}/manifest.json"
        if not os.path.exists(manifest_file):
            raise click.ClickException(
                f"Manifest file not found at '{manifest_file}'. You might need to 'dbt compile'."
            )
        with open(manifest_file, "r") as file:
            manifest = json.load(file)

        sources = _generate_sources(profiles, dbt_target, dbt.prefix)
        models = _generate_models(manifest, dbt_profile, dbt_target, dbt.prefix)

        output_file = dbt.get_output_file(
            output_dir=output_dir, working_dir=working_dir
        )

        output_dir_path = os.path.dirname(output_file)
        if not os.path.exists(output_dir_path):
            os.makedirs(output_dir_path)
        with open(output_file, "w") as file:
            yaml.dump({"sources": sources, "models": models}, file)

    else:
        Logger.instance().debug(f"dbt is not enabled.")



from visivo.models.trace import Trace
import click
import requests
import json
import sys
import asyncio
import aiofiles
import httpx
import contextvars
from time import time
from tenacity import retry, stop_after_attempt, wait_fixed
from visivo.commands.utils import get_profile_file, get_profile_token
from visivo.discovery.discover import Discover
from visivo.logging.logger import Logger
from visivo.parsers.serializer import Serializer
from visivo.parsers.parser_factory import ParserFactory

# Limit concurrent uploads to avoid overloading the API
semaphore = asyncio.Semaphore(50)
attempt = contextvars.ContextVar("attempt")
MAX_ATTEMPTS = 3


@retry(stop=stop_after_attempt(MAX_ATTEMPTS), wait=wait_fixed(2))
async def upload_trace_data(trace, output_dir, form_headers, host, progress):
    """
    Asynchronously uploads trace data files.
    """
    attempt.set(attempt.get(0) + 1)
    async with semaphore:
        try:
            data_file = f"{output_dir}/{trace.name}/data.json"
            async with httpx.AsyncClient(timeout=60) as client:
                async with aiofiles.open(data_file, "rb") as f:
                    files = {
                        "file": (
                            f"{trace.name}.json",
                            await f.read(),
                            "application/json",
                        )
                    }
                    url = f"{host}/api/files/"
                    response = await client.post(url, files=files, headers=form_headers)
                    response.raise_for_status()
                    progress["completed"] += 1
                    Logger.instance().success(
                        f"\tTrace '{trace.name}' data uploaded [{progress['completed']}/{progress['total']}]"
                    )
                    return response.json()["id"]
        except httpx.HTTPStatusError as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] HTTP error while creating trace '{trace.name}': {repr(e)} - Response: {e.response.text}"
            )
            if attempt.get() == MAX_ATTEMPTS:
                progress["failed"].append(f"Data upload for '{trace.name}': {repr(e)}")
            raise
        except Exception as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] Failed to upload trace data for '{trace.name}': {repr(e)}"
            )
            if attempt.get() == MAX_ATTEMPTS:
                progress["failed"].append(f"Data upload for '{trace.name}': {repr(e)}")
            raise


@retry(stop=stop_after_attempt(MAX_ATTEMPTS), wait=wait_fixed(2))
async def create_trace_record(
    trace, project_id, data_file_id, json_headers, host, progress
):
    """
    Asynchronously creates a trace record on the server.
    """
    async with semaphore:
        try:
            body = {
                "name": trace.name,
                "project_id": project_id,
                "data_file_id": data_file_id,
            }
            async with httpx.AsyncClient(timeout=60) as client:
                url = f"{host}/api/traces/"
                response = await client.post(url, json=body, headers=json_headers)
                response.raise_for_status()
                progress["completed"] += 1
                Logger.instance().success(
                    f"\tTrace '{trace.name}' created [{progress['completed']}/{progress['total']}]"
                )
        except httpx.HTTPStatusError as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] HTTP error while creating trace '{trace.name}': {repr(e)} - Response: {e.response.text}"
            )
            if attempt.get() == MAX_ATTEMPTS:
                progress["failed"].append(
                    f"Record creation for '{trace.name}': {repr(e)}"
                )
            raise
        except Exception as e:
            Logger.instance().error(
                f"\t[Attempt {attempt.get()}/{MAX_ATTEMPTS}] Failed to create trace '{trace.name}': {repr(e)}"
            )
            if attempt.get() == MAX_ATTEMPTS:
                progress["failed"].append(
                    f"Record creation for '{trace.name}': {repr(e)}"
                )
            raise


async def process_traces_async(
    traces, output_dir, project_id, form_headers, json_headers, host
):
    """
    Coordinates the asynchronous upload of trace data files and the creation of trace records.
    """
    total_operations = (
        len(traces) * 2
    )  # Each trace has a data upload and record creation
    progress = {"completed": 0, "total": total_operations, "failed": []}

    tasks = []
    for trace in traces:
        # Upload data files concurrently
        data_file_task = upload_trace_data(
            trace, output_dir, form_headers, host, progress
        )
        tasks.append(data_file_task)

    # Wait for all data uploads to complete and gather results
    data_file_ids = await asyncio.gather(*tasks, return_exceptions=True)

    # Prepare to create trace records based on successful uploads
    record_tasks = []
    for trace, data_file_id in zip(traces, data_file_ids):
        if isinstance(data_file_id, Exception):
            Logger.instance().error(
                f"\tSkipping trace '{trace.name}' due to previous error uploading data."
            )
            continue
        record_task = create_trace_record(
            trace, project_id, data_file_id, json_headers, host, progress
        )
        record_tasks.append(record_task)

    # Execute the creation of trace records concurrently
    await asyncio.gather(*record_tasks, return_exceptions=True)

    return progress["failed"]


def deploy_phase(working_dir, user_dir, output_dir, stage, host):
    """
    Synchronous function to manage the deployment, including initiating asynchronous operations.
    """
    deploy_start_time = time()
    # Retrieve profile token for authentication
    Logger.instance().debug("Retrieving profile token...")
    profile_file = get_profile_file(home_dir=user_dir)
    profile_token = get_profile_token(profile_file)
    Logger.instance().info(f"Found Profile token: {profile_file}")

    # Discover and parse project details
    Logger.instance().info("")
    Logger.instance().debug("Compiling project details...")
    discover = Discover(
        working_dir=working_dir, home_dir=user_dir, output_dir=output_dir
    )
    parser = ParserFactory().build(
        project_file=discover.project_file, files=discover.files
    )
    project = parser.parse()
    serializer = Serializer(project=project)
    project_json = json.loads(
        serializer.dereference().model_dump_json(exclude_none=True, by_alias=True)
    )
    Logger.instance().success(
        f"Project Compiled in {time() - deploy_start_time:.2f} seconds"
    )

    # Prepare request payloads and headers
    body = {
        "project_json": project_json,
        "name": project_json["name"],
        "cli_version": project_json["cli_version"],
        "stage": stage,
    }
    json_headers = {
        "content-type": "application/json",
        "Authorization": f"Api-Key {profile_token}",
    }
    form_headers = {
        "Authorization": f"Api-Key {profile_token}",
    }

    # Upload the project information (synchronous)
    Logger.instance().debug("Uploading project information...")
    upload_project_start_time = time()
    url = f"{host}/api/projects/"
    response = requests.post(url, data=json.dumps(body), headers=json_headers)
    if response.status_code == 401:
        raise click.ClickException(f"Token not authorized for host: {host}")
    if response.status_code == 404:
        raise click.ClickException(f"404 error raised. Does your user have an account?")
    if response.status_code == 201:
        Logger.instance().success(
            f"Project uploaded in {time() - upload_project_start_time:.2f} seconds"
        )
        project_data = response.json()
        project_id = project_data["id"]
        project_url = project_data["url"]

        # Async processing of trace uploads and record creations
        Logger.instance().info(f"")
        Logger.instance().info("Processing trace uploads and record creations...")
        process_traces_start_time = time()

        traces = project.descendants_of_type(type=Trace)
        failed_operations = asyncio.run(
            process_traces_async(
                traces=traces,
                output_dir=output_dir,
                project_id=project_id,
                form_headers=form_headers,
                json_headers=json_headers,
                host=host,
            )
        )
        Logger.instance().info(
            f"Trace uploads and record creations completed in {time() - process_traces_start_time:.2f} seconds"
        )

        if failed_operations:
            Logger.instance().info("")
            Logger.instance().info("The following operations failed:")
            for failure in failed_operations:
                Logger.instance().error(f"\t{failure}")
            Logger.instance().info("")
            Logger.instance().info(
                f"Deployment failed in {time() - deploy_start_time:.2f} seconds"
            )
            sys.exit(1)
        Logger.instance().success(
            f"Deployment completed in {time() - deploy_start_time:.2f} seconds"
        )
        return project_url
    else:
        Logger.instance().info(
            f"Deployment failed in {time() - deploy_start_time:.2f} seconds"
        )
        raise click.ClickException(f"There was an unexpected error: {response.content}")




from pydantic import Field, constr
from visivo.models.base.base_model import INDEXED_STATEMENT_REGEX, STATEMENT_REGEX
from visivo.models.trace_props.trace_props import  TraceProps, TracePropsAttribute
from typing import List, Literal, Optional, Any 


class TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br> """
	)
	family: Optional[str]= Field(
		None,
		description=""" string<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 1<br> """
	)
class Tickformatstops1(TracePropsAttribute):
	dtickrange: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" array<br>range ["min", "max"], where "min", "max" - dtick values which describe some zoom level, it is possible to omit "min" or "max" value by passing "null" """
	)
	enabled: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether or not this stop is used. If `false`, this stop is ignored even within its `dtickrange`. """
	)
	name: Optional[str]= Field(
		None,
		description=""" string<br>When used in a template, named items are created in the output figure in addition to any items the figure already has in this array. You can modify these items in the output figure by making your own item with `templateitemname` matching this `name` alongside your modifications (including `visible: false` or `enabled: false` to hide it). Has no effect outside of a template. """
	)
	templateitemname: Optional[str]= Field(
		None,
		description=""" string<br>Used to refer to a named item in this array in the template. Named items from the template will be created even without a matching item in the input figure, but you can modify one by making an item with `templateitemname` matching its `name`, alongside your modifications (including `visible: false` or `enabled: false` to hide it). If there is no template or no matching item, this item will be hidden unless you explicitly show it with `visible: true`. """
	)
	value: Optional[str]= Field(
		None,
		description=""" string<br>string - dtickformat for described zoom level, the same as "tickformat" """
	)
class TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br> """
	)
	family: Optional[str]= Field(
		None,
		description=""" string<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 1<br> """
	)
class ContourcarpetColorbarTitle(TracePropsAttribute):
	font: Optional[TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets this color bar's title font. Note that the title's font used to be set by the now deprecated `titlefont` attribute. """
	)
	side: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "right" | "top" | "bottom" )<br>Determines the location of color bar's title with respect to the color bar. Defaults to "top" when `orientation` if "v" and defaults to "right" when `orientation` if "h". Note that the title's location used to be set by the now deprecated `titleside` attribute. """
	)
	text: Optional[str]= Field(
		None,
		description=""" string<br>Sets the title of the color bar. Note that before the existence of `title.text`, the title's contents used to be defined as the `title` attribute itself. This behavior has been deprecated. """
	)
class ContourcarpetColorbar(TracePropsAttribute):
	bgcolor: Optional[str]= Field(
		None,
		description=""" color<br>Sets the color of padded area. """
	)
	bordercolor: Optional[str]= Field(
		None,
		description=""" color<br>Sets the axis line color. """
	)
	borderwidth: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the width (in px) or the border enclosing this color bar. """
	)
	dtick: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Sets the step in-between ticks on this axis. Use with `tick0`. Must be a positive number, or special strings available to "log" and "date" axes. If the axis `type` is "log", then ticks are set every 10^(n"dtick) where n is the tick number. For example, to set a tick mark at 1, 10, 100, 1000, ... set dtick to 1. To set tick marks at 1, 100, 10000, ... set dtick to 2. To set tick marks at 1, 5, 25, 125, 625, 3125, ... set dtick to log_10(5), or 0.69897000433. "log" has several special values; "L<f>", where `f` is a positive number, gives ticks linearly spaced in value (but not position). For example `tick0` = 0.1, `dtick` = "L0.5" will put ticks at 0.1, 0.6, 1.1, 1.6 etc. To show powers of 10 plus small digits between, use "D1" (all digits) or "D2" (only 2 and 5). `tick0` is ignored for "D1" and "D2". If the axis `type` is "date", then you must convert the time to milliseconds. For example, to set the interval between ticks to one day, set `dtick` to 86400000.0. "date" also has special values "M<n>" gives ticks spaced by a number of months. `n` must be a positive integer. To set ticks on the 15th of every third month, set `tick0` to "2000-01-15" and `dtick` to "M3". To set ticks every 4 years, set `dtick` to "M48" """
	)
	exponentformat: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "none" | "e" | "e" | "power" | "si" | "b" )<br>Determines a formatting rule for the tick exponents. For example, consider the number 1,000,000,000. If "none", it appears as 1,000,000,000. If "e", 1e+9. If "E", 1E+9. If "power", 1x10^9 (with 9 in a super script). If "SI", 1G. If "B", 1B. """
	)
	len: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the length of the color bar This measure excludes the padding of both ends. That is, the color bar length is this length minus the padding on both ends. """
	)
	lenmode: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "fraction" | "pixels" )<br>Determines whether this color bar's length (i.e. the measure in the color variation direction) is set in units of plot "fraction" or in "pixels. Use `len` to set the value. """
	)
	minexponent: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Hide SI prefix for 10^n if |n| is below this number. This only has an effect when `tickformat` is "SI" or "B". """
	)
	nticks: Optional[int | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" integer greater than or equal to 0<br>Specifies the maximum number of ticks for the particular axis. The actual number of ticks will be chosen automatically to be less than or equal to `nticks`. Has an effect only if `tickmode` is set to "auto". """
	)
	orientation: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "h" | "v" )<br>Sets the orientation of the colorbar. """
	)
	outlinecolor: Optional[str]= Field(
		None,
		description=""" color<br>Sets the axis line color. """
	)
	outlinewidth: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the width (in px) of the axis line. """
	)
	separatethousands: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>If "true", even 4-digit integers are separated """
	)
	showexponent: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "all" | "first" | "last" | "none" )<br>If "all", all exponents are shown besides their significands. If "first", only the exponent of the first tick is shown. If "last", only the exponent of the last tick is shown. If "none", no exponents appear. """
	)
	showticklabels: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether or not the tick labels are drawn. """
	)
	showtickprefix: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "all" | "first" | "last" | "none" )<br>If "all", all tick labels are displayed with a prefix. If "first", only the first tick is displayed with a prefix. If "last", only the last tick is displayed with a suffix. If "none", tick prefixes are hidden. """
	)
	showticksuffix: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "all" | "first" | "last" | "none" )<br>Same as `showtickprefix` but for tick suffixes. """
	)
	thickness: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the thickness of the color bar This measure excludes the size of the padding, ticks and labels. """
	)
	thicknessmode: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "fraction" | "pixels" )<br>Determines whether this color bar's thickness (i.e. the measure in the constant color direction) is set in units of plot "fraction" or in "pixels". Use `thickness` to set the value. """
	)
	tick0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Sets the placement of the first tick on this axis. Use with `dtick`. If the axis `type` is "log", then you must take the log of your starting tick (e.g. to set the starting tick to 100, set the `tick0` to 2) except when `dtick`="L<f>" (see `dtick` for more info). If the axis `type` is "date", it should be a date string, like date data. If the axis `type` is "category", it should be a number, using the scale where each category is assigned a serial number from zero in the order it appears. """
	)
	tickangle: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" angle<br>Sets the angle of the tick labels with respect to the horizontal. For example, a `tickangle` of -90 draws the tick labels vertically. """
	)
	tickcolor: Optional[str]= Field(
		None,
		description=""" color<br>Sets the tick color. """
	)
	tickfont: Optional[TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets the color bar's tick label font """
	)
	tickformat: Optional[str]= Field(
		None,
		description=""" string<br>Sets the tick label formatting rule using d3 formatting mini-languages which are very similar to those in Python. For numbers, see: https://github.com/d3/d3-format/tree/v1.4.5#d3-format. And for dates see: https://github.com/d3/d3-time-format/tree/v2.2.3#locale_format. We add two items to d3's date formatter: "%h" for half of the year as a decimal number as well as "%{n}f" for fractional seconds with n digits. For example, "2016-10-13 09:15:23.456" with tickformat "%H~%M~%S.%2f" would display "09~15~23.46" """
	)
	tickformatstops: Optional[List[Tickformatstops1]]= Field(
		None,
		description=""" array of object where each object has one or more of the keys listed below.<br> """
	)
	ticklabeloverflow: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "allow" | "hide past div" | "hide past domain" )<br>Determines how we handle tick labels that would overflow either the graph div or the domain of the axis. The default value for inside tick labels is "hide past domain". In other cases the default is "hide past div". """
	)
	ticklabelposition: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "outside" | "inside" | "outside top" | "inside top" | "outside left" | "inside left" | "outside right" | "inside right" | "outside bottom" | "inside bottom" )<br>Determines where tick labels are drawn relative to the ticks. Left and right options are used when `orientation` is "h", top and bottom when `orientation` is "v". """
	)
	ticklabelstep: Optional[int | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" integer greater than or equal to 1<br>Sets the spacing between tick labels as compared to the spacing between ticks. A value of 1 (default) means each tick gets a label. A value of 2 means shows every 2nd label. A larger value n means only every nth tick is labeled. `tick0` determines which labels are shown. Not implemented for axes with `type` "log" or "multicategory", or when `tickmode` is "array". """
	)
	ticklen: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the tick length (in px). """
	)
	tickmode: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "auto" | "linear" | "array" )<br>Sets the tick mode for this axis. If "auto", the number of ticks is set via `nticks`. If "linear", the placement of the ticks is determined by a starting position `tick0` and a tick step `dtick` ("linear" is the default value if `tick0` and `dtick` are provided). If "array", the placement of the ticks is set via `tickvals` and the tick text is `ticktext`. ("array" is the default value if `tickvals` is provided). """
	)
	tickprefix: Optional[str]= Field(
		None,
		description=""" string<br>Sets a tick label prefix. """
	)
	ticks: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "outside" | "inside" | "" )<br>Determines whether ticks are drawn or not. If "", this axis' ticks are not drawn. If "outside" ("inside"), this axis' are drawn outside (inside) the axis lines. """
	)
	ticksuffix: Optional[str]= Field(
		None,
		description=""" string<br>Sets a tick label suffix. """
	)
	ticktext: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the text displayed at the ticks position via `tickvals`. Only has an effect if `tickmode` is set to "array". Used with `tickvals`. """
	)
	tickvals: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the values at which ticks on this axis appear. Only has an effect if `tickmode` is set to "array". Used with `ticktext`. """
	)
	tickwidth: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the tick width (in px). """
	)
	title: Optional[ContourcarpetColorbarTitle]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	x: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to -2 and 3<br>Sets the x position of the color bar (in plot fraction). Defaults to 1.02 when `orientation` is "v" and 0.5 when `orientation` is "h". """
	)
	xanchor: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "left" | "center" | "right" )<br>Sets this color bar's horizontal position anchor. This anchor binds the `x` position to the "left", "center" or "right" of the color bar. Defaults to "left" when `orientation` is "v" and "center" when `orientation` is "h". """
	)
	xpad: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the amount of padding (in px) along the x direction. """
	)
	y: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to -2 and 3<br>Sets the y position of the color bar (in plot fraction). Defaults to 0.5 when `orientation` is "v" and 1.02 when `orientation` is "h". """
	)
	yanchor: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "top" | "middle" | "bottom" )<br>Sets this color bar's vertical position anchor This anchor binds the `y` position to the "top", "middle" or "bottom" of the color bar. Defaults to "middle" when `orientation` is "v" and "bottom" when `orientation` is "h". """
	)
	ypad: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the amount of padding (in px) along the y direction. """
	)
class TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br> """
	)
	family: Optional[str]= Field(
		None,
		description=""" string<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 1<br> """
	)
class ContourcarpetContours(TracePropsAttribute):
	coloring: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "fill" | "lines" | "none" )<br>Determines the coloring method showing the contour values. If "fill", coloring is done evenly between each contour level If "lines", coloring is done on the contour lines. If "none", no coloring is applied on this trace. """
	)
	end: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the end contour level value. Must be more than `contours.start` """
	)
	labelfont: Optional[TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets the font used for labeling the contour levels. The default color comes from the lines, if shown. The default family and size come from `layout.font`. """
	)
	labelformat: Optional[str]= Field(
		None,
		description=""" string<br>Sets the contour label formatting rule using d3 formatting mini-languages which are very similar to those in Python. For numbers, see: https://github.com/d3/d3-format/tree/v1.4.5#d3-format. """
	)
	operation: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "=" | "&lt;" | "&gt;=" | "&gt;" | "&lt;=" | "[]" | "()" | "[)" | "(]" | "][" | ")(" | "](" | ")[" )<br>Sets the constraint operation. "=" keeps regions equal to `value` "<" and "<=" keep regions less than `value` ">" and ">=" keep regions greater than `value` "[]", "()", "[)", and "(]" keep regions inside `value[0]` to `value[1]` "][", ")(", "](", ")[" keep regions outside `value[0]` to value[1]` Open vs. closed intervals make no difference to constraint display, but all versions are allowed for consistency with filter transforms. """
	)
	showlabels: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether to label the contour lines with their values. """
	)
	showlines: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether or not the contour lines are drawn. Has an effect only if `contours.coloring` is set to "fill". """
	)
	size: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the step between each contour level. Must be positive. """
	)
	start: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the starting contour level value. Must be less than `contours.end` """
	)
	type: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "levels" | "constraint" )<br>If `levels`, the data is represented as a contour plot with multiple levels displayed. If `constraint`, the data is represented as constraints with the invalid region shaded as specified by the `operation` and `value` parameters. """
	)
	value: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Sets the value or values of the constraint boundary. When `operation` is set to one of the comparison values (=,<,>=,>,<=) "value" is expected to be a number. When `operation` is set to one of the interval values ([],(),[),(],][,)(,](,)[) "value" is expected to be an array of two numbers where the first is the lower bound and the second is the upper bound. """
	)
class TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br> """
	)
	family: Optional[str]= Field(
		None,
		description=""" string<br>HTML font family - the typeface that will be applied by the web browser. The web browser will only be able to apply a font if it is available on the system which it operates. Provide multiple font families, separated by commas, to indicate the preference in which to apply fonts if they aren't available on the system. The Chart Studio Cloud (at https://chart-studio.plotly.com or on-premise) generates images on a server, where only a select number of fonts are installed and supported. These include "Arial", "Balto", "Courier New", "Droid Sans",, "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman". """
	)
	size: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 1<br> """
	)
class ContourcarpetLegendgrouptitle(TracePropsAttribute):
	font: Optional[TextfontLabelfontTickfontInsidetextfontRangefontOutsidetextfontFontGrouptitlefont1]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br>Sets this legend group's title font. """
	)
	text: Optional[str]= Field(
		None,
		description=""" string<br>Sets the title of the legend group. """
	)
class Line7(TracePropsAttribute):
	color: Optional[str]= Field(
		None,
		description=""" color<br>Sets the color of the contour level. Has no effect if `contours.coloring` is set to "lines". """
	)
	dash: Optional[str]= Field(
		None,
		description=""" string<br>Sets the dash style of lines. Set to a dash type string ("solid", "dot", "dash", "longdash", "dashdot", or "longdashdot") or a dash length list in px (eg "5px,10px,2px,2px"). """
	)
	smoothing: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to 0 and 1.3<br>Sets the amount of smoothing for the contour lines, where "0" corresponds to no smoothing. """
	)
	width: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the contour line width in (in px) Defaults to "0.5" when `contours.type` is "levels". Defaults to "2" when `contour.type` is "constraint". """
	)
class Contourcarpet(TraceProps):
	a0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Alternate to `x`. Builds a linear space of x coordinates. Use with `dx` where `x0` is the starting coordinate and `dx` the step. """
	)
	a: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the x coordinates. """
	)
	atype: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "array" | "scaled" )<br>If "array", the heatmap's x coordinates are given by "x" (the default behavior when `x` is provided). If "scaled", the heatmap's x coordinates are given by "x0" and "dx" (the default behavior when `x` is not provided). """
	)
	autocolorscale: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether the colorscale is a default palette (`autocolorscale: true`) or the palette determined by `colorscale`. In case `colorscale` is unspecified or `autocolorscale` is true, the default palette will be chosen according to whether numbers in the `color` array are all positive, all negative or mixed. """
	)
	autocontour: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether or not the contour level attributes are picked by an algorithm. If "true", the number of contour levels can be set in `ncontours`. If "false", set the contour level attributes in `contours`. """
	)
	b0: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Alternate to `y`. Builds a linear space of y coordinates. Use with `dy` where `y0` is the starting coordinate and `dy` the step. """
	)
	b: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the y coordinates. """
	)
	btype: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( "array" | "scaled" )<br>If "array", the heatmap's y coordinates are given by "y" (the default behavior when `y` is provided) If "scaled", the heatmap's y coordinates are given by "y0" and "dy" (the default behavior when `y` is not provided) """
	)
	carpet: Optional[str]= Field(
		None,
		description=""" string<br>The `carpet` of the carpet axes on which this contour trace lies """
	)
	coloraxis: Optional[str]= Field(
		None,
		description=""" subplotid<br>Sets a reference to a shared color axis. References to these shared color axes are "coloraxis", "coloraxis2", "coloraxis3", etc. Settings for these shared color axes are set in the layout, under `layout.coloraxis`, `layout.coloraxis2`, etc. Note that multiple color scales can be linked to the same color axis. """
	)
	colorbar: Optional[ContourcarpetColorbar]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	colorscale: Optional[Any]= Field(
		None,
		description=""" colorscale<br>Sets the colorscale. The colorscale must be an array containing arrays mapping a normalized value to an rgb, rgba, hex, hsl, hsv, or named color string. At minimum, a mapping for the lowest (0) and highest (1) values are required. For example, `[[0, 'rgb(0,0,255)'], [1, 'rgb(255,0,0)']]`. To control the bounds of the colorscale in color space, use `zmin` and `zmax`. Alternatively, `colorscale` may be a palette name string of the following list: Blackbody,Bluered,Blues,Cividis,Earth,Electric,Greens,Greys,Hot,Jet,Picnic,Portland,Rainbow,RdBu,Reds,Viridis,YlGnBu,YlOrRd. """
	)
	contours: Optional[ContourcarpetContours]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	customdata: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Assigns extra data each datum. This may be useful when listening to hover, click and selection events. Note that, "scatter" traces also appends customdata items in the markers DOM elements """
	)
	da: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the x coordinate step. See `x0` for more info. """
	)
	db: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the y coordinate step. See `y0` for more info. """
	)
	fillcolor: Optional[str]= Field(
		None,
		description=""" color<br>Sets the fill color if `contours.type` is "constraint". Defaults to a half-transparent variant of the line color, marker color, or marker line color, whichever is available. """
	)
	hovertext: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Same as `text`. """
	)
	ids: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Assigns id labels to each datum. These ids for object constancy of data points during animation. Should be an array of strings, not numbers or any other type. """
	)
	legendgroup: Optional[str]= Field(
		None,
		description=""" string<br>Sets the legend group for this trace. Traces part of the same legend group hide/show at the same time when toggling legend items. """
	)
	legendgrouptitle: Optional[ContourcarpetLegendgrouptitle]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	legendrank: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the legend rank for this trace. Items and groups with smaller ranks are presented on top/left side while with `"reversed" `legend.traceorder` they are on bottom/right side. The default legendrank is 1000, so that you can use ranks less than 1000 to place certain items before all unranked items, and ranks greater than 1000 to go after all unranked items. """
	)
	legendwidth: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number greater than or equal to 0<br>Sets the width (in px or fraction) of the legend for this trace. """
	)
	line: Optional[Line7]= Field(
		None,
		description=""" object containing one or more of the keys listed below.<br> """
	)
	meta: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Assigns extra meta information associated with this trace that can be used in various text attributes. Attributes such as trace `name`, graph, axis and colorbar `title.text`, annotation `text` `rangeselector`, `updatemenues` and `sliders` `label` text all support `meta`. To access the trace `meta` values in an attribute in the same trace, simply use `%{meta[i]}` where `i` is the index or key of the `meta` item in question. To access trace `meta` in layout attributes, use `%{data[n[.meta[i]}` where `i` is the index or key of the `meta` and `n` is the trace index. """
	)
	ncontours: Optional[int | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" integer greater than or equal to 1<br>Sets the maximum number of contour levels. The actual number of contours will be chosen automatically to be less than or equal to the value of `ncontours`. Has an effect only if `autocontour` is "true" or if `contours.size` is missing. """
	)
	opacity: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number between or equal to 0 and 1<br>Sets the opacity of the trace. """
	)
	reversescale: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Reverses the color mapping if true. If true, `zmin` will correspond to the last color in the array and `zmax` will correspond to the first color. """
	)
	showlegend: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether or not an item corresponding to this trace is shown in the legend. """
	)
	showscale: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether or not a colorbar is displayed for this trace. """
	)
	text: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the text elements associated with each z value. """
	)
	transpose: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Transposes the z data. """
	)
	type: Literal["contourcarpet"]= Field(
		...,
		description=""" "contourcarpet"<br> """
	)
	uirevision: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number or categorical coordinate string<br>Controls persistence of some user-driven changes to the trace: `constraintrange` in `parcoords` traces, as well as some `editable: true` modifications such as `name` and `colorbar.title`. Defaults to `layout.uirevision`. Note that other user-driven trace attribute changes are controlled by `layout` attributes: `trace.visible` is controlled by `layout.legend.uirevision`, `selectedpoints` is controlled by `layout.selectionrevision`, and `colorbar.(x|y)` (accessible with `config: {editable: true}`) is controlled by `layout.editrevision`. Trace changes are tracked by `uid`, which only falls back on trace index if no `uid` is provided. So if your app can add/remove traces before the end of the `data` array, such that the same trace has a different index, you can still preserve user-driven changes if you give each trace a `uid` that stays with it as it moves. """
	)
	visible: Optional[str]= Field(
		None,
		description=""" enumerated , one of ( true | false | "legendonly" )<br>Determines whether or not this trace is visible. If "legendonly", the trace is not drawn, but can appear as a legend item (provided that the legend itself is visible). """
	)
	xaxis: Optional[str]= Field(
		None,
		description=""" subplotid<br>Sets a reference between this trace's x coordinates and a 2D cartesian x axis. If "x" (the default value), the x coordinates refer to `layout.xaxis`. If "x2", the x coordinates refer to `layout.xaxis2`, and so on. """
	)
	yaxis: Optional[str]= Field(
		None,
		description=""" subplotid<br>Sets a reference between this trace's y coordinates and a 2D cartesian y axis. If "y" (the default value), the y coordinates refer to `layout.yaxis`. If "y2", the y coordinates refer to `layout.yaxis2`, and so on. """
	)
	z: Optional[constr(pattern=STATEMENT_REGEX) | List]= Field(
		None,
		description=""" data array<br>Sets the z data. """
	)
	zauto: Optional[bool | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" boolean<br>Determines whether or not the color domain is computed with respect to the input data (here in `z`) or the bounds set in `zmin` and `zmax` Defaults to `false` when `zmin` and `zmax` are set by the user. """
	)
	zmax: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the upper bound of the color domain. Value should have the same units as in `z` and if set, `zmin` must be set as well. """
	)
	zmid: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the mid-point of the color domain by scaling `zmin` and/or `zmax` to be equidistant to this point. Value should have the same units as in `z`. Has no effect when `zauto` is `false`. """
	)
	zmin: Optional[float | constr(pattern=INDEXED_STATEMENT_REGEX)]= Field(
		None,
		description=""" number<br>Sets the lower bound of the color domain. Value should have the same units as in `z` and if set, `zmax` must be set as well. """
	)


import sys
from visivo.logging.logger import Logger
from visivo.models.test import Test
from visivo.testing.runner import Runner
from visivo.commands.compile_phase import compile_phase


def test_phase(output_dir: str, default_source: str, working_dir: str):
    project = compile_phase(
        default_source=default_source, working_dir=working_dir, output_dir=output_dir
    )
    Logger.instance().debug("Testing project")

    dag = project.dag()
    tests = project.descendants_of_type(type=Test)

    test_runner = Runner(
        tests=tests,
        project=project,
        output_dir=output_dir,
        dag=dag,
    )
    if not test_runner.run().success:
        sys.exit(1)


test_phase.__test__ = False



from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import smtplib
from typing import Literal

from pydantic import Field

from visivo.models.destinations.destination import Destination
from visivo.models.test_run import TestRun


class EmailDestination(Destination):
    """
    You can configure email destinations for any SMTP provider. Here's an example of this configuration looks in your yaml file:
    ``` yaml
    destinations:
      - name: email-destination #any unique name of your choosing
        type: email
        subject: "[ALERT] Your Visivo Tests Have Failed" #can be any message you want
        to: someone@your_company.com
        port: 2525 #is this port by default
        host: your_company_email_server.com
        username: someones_username
        password: {% raw %}{{ env_var('EMAIL_PASSWORD')}}{% endraw %} #We'd recommend using environment variables here for security

    ```
    """

    type: Literal["email"] = Field(None, description="The type of alert destination.")
    subject: str = Field(
        "Visivo Destination", description="Subject of the alert email."
    )
    to: str = Field(None, description="The email to send the alert to.")
    port: int = Field(
        2525,
        description="The port of the email server that the destination is connecting to.",
    )
    host: str = Field(
        None,
        description="The host of the email server that the destination is connecting to.",
    )
    username: str = Field(
        None, description="The username for authenticating the email server."
    )
    password: str = Field(
        None, description="The password for authenticating the email server."
    )

    def alert(self, test_run: TestRun):
        if not test_run.failures:
            return

        body = f"There were test failures running against {test_run.target_name}\n\n"

        for test_failure in test_run.failures:
            body += f"* {test_failure.test_id} - {test_failure.message}\n"

        sender_email = "alerts@visio.io"

        message = MIMEMultipart("alternative")
        message["Subject"] = self.subject
        message["From"] = sender_email
        message["To"] = self.to

        html = f"""\
        <html>
        <body>
            {body}
        </body>
        </html>
        """

        part1 = MIMEText(body, "plain")
        part2 = MIMEText(html, "html")
        message.attach(part1)
        message.attach(part2)

        with smtplib.SMTP(self.host, self.port) as server:
            server.login(self.login, self.password)
            server.sendmail(sender_email, self.to, message.as_string())



import re
from typing import List, Optional
from visivo.models.alert import Alert
from visivo.models.dag import all_descendants_of_type
from visivo.models.destinations.destination import Destination

from visivo.models.destinations.fields import DestinationField
from visivo.models.include import Include
from visivo.models.models.model import Model
from visivo.models.models.fields import ModelField
from visivo.models.models.sql_model import SqlModel
from visivo.models.selector import Selector, SelectorType
from visivo.models.sources.fields import SourceField


from .base.parent_model import ParentModel
from .dashboard import Dashboard
from .chart import Chart
from .trace import Trace
from .sources.source import Source
from .table import Table
from .defaults import Defaults
from .dbt import Dbt
from typing import List
from .base.named_model import NamedModel
from .base.base_model import BaseModel
from pydantic import ConfigDict, Field, model_validator


class Project(NamedModel, ParentModel):
    model_config = ConfigDict(populate_by_name=True, extra="forbid")

    defaults: Optional[Defaults] = None
    dbt: Optional[Dbt] = None
    cli_version: Optional[str] = None
    includes: List[Include] = []
    destinations: List[DestinationField] = []
    alerts: List[Alert] = []
    sources: List[SourceField] = Field(
        [],
        description="A list of source objects.",
        alias="targets",
    )
    models: List[ModelField] = []
    traces: List[Trace] = []
    tables: List[Table] = []
    charts: List[Chart] = []
    selectors: List[Selector] = []
    dashboards: List[Dashboard] = []

    def child_items(self):
        return (
            self.destinations
            + self.alerts
            + self.sources
            + self.models
            + self.traces
            + self.tables
            + self.charts
            + self.selectors
            + self.dashboards
        )

    def filter_traces(self, name_filter) -> List[Trace]:
        if name_filter:
            included_nodes = self.nodes_including_named_node_in_graph(name=name_filter)
        else:
            included_nodes = self.descendants()
        return set(self.descendants_of_type(Trace)).intersection(included_nodes)

    @model_validator(mode="after")
    def validate_default_names(self):
        sources, alerts = (self.sources, self.alerts)
        source_names = [source.name for source in sources]
        alert_names = [alert.name for alert in alerts]
        defaults = self.defaults
        if not defaults:
            return self

        if defaults.source_name and defaults.source_name not in source_names:
            raise ValueError(f"default source '{defaults.source_name}' does not exist")

        if defaults.alert_name and defaults.alert_name not in alert_names:
            raise ValueError(f"default alert '{defaults.alert_name}' does not exist")

        return self

    @model_validator(mode="after")
    def validate_models_have_sources(self):
        defaults = self.defaults
        if defaults and defaults.source_name:
            return self

        for model in self.descendants_of_type(Model):
            if isinstance(model, SqlModel) and not model.source:
                raise ValueError(
                    f"'{model.name}' does not specify a source and project does not specify default source"
                )

        return self

    @model_validator(mode="after")
    def validate_dag(self):
        dag = self.dag()
        tables = all_descendants_of_type(type=Table, dag=dag, from_node=self)
        for table in tables:
            selectors = all_descendants_of_type(type=Selector, dag=dag, from_node=table)
            if len(selectors) > 0 and selectors[0].type == SelectorType.multiple:
                raise ValueError(
                    f"Table with name '{table.name}' has a selector with a 'multiple' type.  This is not permitted."
                )

        return self

    @model_validator(mode="before")
    def set_paths_on_models(cls, values):
        def set_path_recursively(obj, path=""):
            if isinstance(obj, dict):
                obj["path"] = path
                for key, value in obj.items():
                    if key not in ["props", "defaults", "layout", "columns"]:
                        new_path = f"{path}.{key}" if path else key
                        set_path_recursively(value, new_path)
            elif isinstance(obj, list):
                for index, item in enumerate(obj):
                    new_path = f"{path}[{index}]"
                    set_path_recursively(item, new_path)

        set_path_recursively(values, "project")
        return values

    @model_validator(mode="after")
    def validate_names(self):
        Project.traverse_names([], self)
        return self

    @model_validator(mode="before")
    def set_path_on_named_models(cls, values):
        def set_path_recursively(obj, path=""):
            if isinstance(obj, dict):
                obj["path"] = path
                for key, value in obj.items():
                    if key not in ["props", "defaults", "layout", "columns"]:
                        new_path = f"{path}.{key}" if path else key
                        set_path_recursively(value, new_path)
            elif isinstance(obj, list):
                for index, item in enumerate(obj):
                    new_path = f"{path}[{index}]"
                    set_path_recursively(item, new_path)

        set_path_recursively(values, "project")
        return values

    @classmethod
    def traverse_names(cls, names, object):
        if isinstance(object, ParentModel):
            for child_item in object.child_items():
                if isinstance(child_item, BaseModel) and hasattr(child_item, "name"):
                    name = NamedModel.get_name(obj=child_item)
                    if name in names:
                        raise ValueError(
                            f"{child_item.__class__.__name__} name '{name}' is not unique in the project"
                        )
                    if name:
                        names.append(name)
                Project.traverse_names(names, child_item)



def show_dag_fig(dag):
    import plotly.graph_objects as go

    from networkx import random_layout

    pos = random_layout(dag)
    edge_x = []
    edge_y = []
    edge_annotations = []

    for edge in dag.edges():
        x0, y0 = pos[edge[0]]
        x1, y1 = pos[edge[1]]
        edge_x.append(x0)
        edge_x.append(x1)
        edge_x.append(None)
        edge_y.append(y0)
        edge_y.append(y1)
        edge_y.append(None)
        edge_annotations.append(
            dict(
                ax=x0,
                ay=y0,
                axref="x",
                ayref="y",
                x=x1,
                y=y1,
                xref="x",
                yref="y",
                showarrow=True,
                arrowhead=2,
                arrowsize=1,
                arrowwidth=1,
                arrowcolor="#888",
            )
        )

    edge_trace = go.Scatter(
        x=edge_x,
        y=edge_y,
        line=dict(width=0.5, color="#888"),
        hoverinfo="none",
        mode="lines",
    )

    node_x = []
    node_y = []
    node_text = []
    for node in dag.nodes():
        x, y = pos[node]
        node_x.append(x)
        node_y.append(y)
        node_text.append(str(node))

    node_trace = go.Scatter(
        x=node_x,
        y=node_y,
        mode="markers+text",
        hoverinfo="text",
        text=node_text,
        marker=dict(
            showscale=True,
            colorscale="YlGnBu",
            size=10,
            colorbar=dict(
                thickness=15,
                title="Node Connections",
                xanchor="left",
                titleside="right",
            ),
        ),
    )

    fig = go.Figure(
        data=[edge_trace, node_trace],
        layout=go.Layout(
            showlegend=False,
            hovermode="closest",
            margin=dict(b=20, l=5, r=5, t=40),
            xaxis=dict(showgrid=False, zeroline=False),
            yaxis=dict(showgrid=False, zeroline=False),
            annotations=edge_annotations,
        ),
    )
    fig.show()


def create_dag_dict(dag):
    nodes = []
    edges = []

    for node in dag.nodes():
        node_data = {
            "id": str(id(node)),
            "name": node.name,
            "path": node.path,
            "type": type(node).__name__.lower(),
        }
        nodes.append(node_data)

    for edge in dag.edges():
        edge_data = {"source": str(id(edge[0])), "target": str(id(edge[1]))}
        edges.append(edge_data)

    return {"nodes": nodes, "edges": edges}


def all_descendants(dag, from_node=None, depth=None):
    import networkx.algorithms.traversal.depth_first_search as dfs

    return dfs.dfs_tree(dag, from_node, depth_limit=depth)


def all_descendants_of_type(type, dag, from_node=None, depth=None):
    if not depth:
        depth = len(dag)

    def find_type(item):
        return isinstance(item, type)

    return list(
        filter(
            find_type,
            all_descendants(dag=dag, from_node=from_node, depth=depth),
        )
    )


def all_descendants_with_name(name: str, dag, from_node=None):
    def find_name(item):
        return hasattr(item, "name") and item.name == name

    return list(filter(find_name, all_descendants(dag=dag, from_node=from_node)))


def all_descendants_with_path_match(path: str, dag, from_node=None):
    def path_match(item):
        return hasattr(item, "path") and item.path and item.path in path

    return list(filter(path_match, all_descendants(dag=dag, from_node=from_node)))


def all_nodes_including_named_node_in_graph(name: str, dag):
    from networkx import descendants, ancestors
    import click

    item = all_descendants_with_name(name=name, dag=dag)

    if len(item) == 1:
        item = item[0]
    else:
        raise click.ClickException(f"No item found with name: '{name}'.")

    d = descendants(dag, item)
    a = ancestors(dag, item)
    items = d.union(a)
    items.add(item)
    return items
